package com.vernaio.example;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import com.vernaio.aivis.engine.v2.base.sdk.EngineLogger;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoFilesFilter;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoInferenceTimeseriesDataFilter;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoInterval;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoTimeseriesDataTypesConfig;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoTimeseriesFilesReaderConfig;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoInferenceDataSpecification;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoSegmentsFloatDataPoint;
import com.vernaio.aivis.engine.v2.state.detection.sdk.StateDetectionAnalysis;
import com.vernaio.aivis.engine.v2.state.detection.sdk.StateDetectionData;
import com.vernaio.aivis.engine.v2.state.detection.sdk.StateDetectionInference;
import com.vernaio.aivis.engine.v2.state.detection.sdk.StateDetectionSetup;
import com.vernaio.aivis.engine.v2.state.detection.sdk.StateDetectionTraining;
import com.vernaio.aivis.engine.v2.state.detection.sdk.dto.DtoAnalysisConfig;
import com.vernaio.aivis.engine.v2.state.detection.sdk.dto.DtoAnalysisSamplingConfig;
import com.vernaio.aivis.engine.v2.state.detection.sdk.dto.DtoInferenceConfig;
import com.vernaio.aivis.engine.v2.state.detection.sdk.dto.DtoSelfLabelingConfig;
import com.vernaio.aivis.engine.v2.state.detection.sdk.dto.DtoTargetConfig;
import com.vernaio.aivis.engine.v2.state.detection.sdk.dto.DtoTrainingConfig;
import com.vernaio.aivis.engine.v2.state.detection.sdk.dto.DtoTrainingSamplingConfig;
import com.vernaio.aivis.engine.v2.state.detection.sdk.dto.IDtoTrainingPreparation;

public class Main {
	
	private static final int MIN_LOGGING_LEVEL = 3;
	
	private static final int MAX_THREADS = 0; // max threads 0 means the number of threads will be chosen automatically
	
	private static final String TRAINING_DATA_FILE = "train_sd.csv";
	
	private static final String INFERENCE_DATA_FILE = "eval_sd.csv";
	
	private static final String TARGET_SIGNAL_ID = "failure";
	
	public static void main(final String[] args) throws Exception {
		
		System.out.println("Setup ...");
		// parsing args as folder config
		final FolderConfig folderConfig = FolderConfig.fromArgs(args);
		// create and register logger
		StateDetectionSetup.registerLogger(new EngineLogger() {
			
			@Override
			public void log(int level, String thread, String module, String message) {
				if (level <= MIN_LOGGING_LEVEL) {
					System.out.println(String.format("\t... %s", message));
				}
			}
		});
		// init thread count
		StateDetectionSetup.initThreadCount(MAX_THREADS);
		System.out.println("\t... done!");
		
		/*
		 * ANALYSIS
		 */
		
		System.out.println("Loading training data ...");
		
		final String[] trainingFiles = { TRAINING_DATA_FILE };
		final DtoFilesFilter trainingFilesFilter = new DtoFilesFilter().withIncludeFiles(trainingFiles);
		
		final String[] booleanSignals = { TARGET_SIGNAL_ID };
		final DtoTimeseriesDataTypesConfig dataTypes = new DtoTimeseriesDataTypesConfig().withBooleanSignals(booleanSignals);
		
		final DtoTimeseriesFilesReaderConfig trainingFilesReaderConfig =
			new DtoTimeseriesFilesReaderConfig(folderConfig.getInputFolder()).withFilesFilter(trainingFilesFilter).withDataTypes(dataTypes);
		
		final StateDetectionData trainingData = StateDetectionData.create();
		trainingData.readFiles(trainingFilesReaderConfig);
		
		System.out.println("Creating analysis ...");
		// build analysis config - for more options here, see documentation
		final DtoAnalysisConfig analysisConfig =
			new DtoAnalysisConfig(
				new DtoTargetConfig(TARGET_SIGNAL_ID), // an incident is, if the target signal switches from false to true
				new DtoSelfLabelingConfig(432000000L, 3600000L), // coherence period: how far in the past before an incident is aivis supposed to look for state changes
				new DtoAnalysisSamplingConfig()
			);
		// create and perform analysis
		final StateDetectionAnalysis analysis = StateDetectionAnalysis.create(trainingData, analysisConfig);
		System.out.println("\t... done!");
		
		System.out.println("Retrieving and saving analysis files ...");
		// training preparation channels information from analysis to training
		final IDtoTrainingPreparation preparation = analysis.getTrainingPreparation();
		// write a json file with information about the analysis
		Utils.writeJson(new File(folderConfig.getOutputFolder(), "analysis-report.json"), analysis.getReport());
		System.out.println("\t... done!");
		
		/*
		 * TRAINING
		 */
		
		System.out.println("Creating training ...");
		// build training config - for more options here, see documentation
		final DtoTrainingConfig trainingConfig =
			new DtoTrainingConfig(preparation.getFunctions(), preparation.getCommonLabelHistory(), preparation.getZoomLevels()[0].getSegments()).withSampling(
				new DtoTrainingSamplingConfig().withMaximalSampleCount(500)
			);
		// create training and train the model
		final StateDetectionTraining training = StateDetectionTraining.create(trainingData, trainingConfig);
		System.out.println("\t... done!");
		
		System.out.println("Retrieving and saving training files ...");
		// model is needed for inference and also contains the inference data specification
		Utils.writeJson(new File(folderConfig.getOutputFolder(), "model.json"), training.getModel());
		Utils.writeJson(new File(folderConfig.getOutputFolder(), "training-report.json"), training.getReport());
		System.out.println("\t... done!");
		
		// saving memory, we don't need training data anymore
		trainingData.destroy();
		
		/*
		 * INFERENCE
		 */
		
		final long inferenceStartTime = 1328472000000L;
		final long inferenceEndTime = 1336572000000L;
		
		System.out.println("Loading inference data ...");
		
		final String[] inferenceFiles = { INFERENCE_DATA_FILE };
		final DtoFilesFilter inferenceFilesFilter = new DtoFilesFilter().withIncludeFiles(inferenceFiles);
		
		final IDtoInferenceDataSpecification dataSpec = training.getModel().getInferenceDataSpecification();
		final DtoInferenceTimeseriesDataFilter inferenceDataFilter =
			new DtoInferenceTimeseriesDataFilter(
				new IDtoInferenceDataSpecification[] { dataSpec },
				new DtoInterval[] { new DtoInterval().withStartTime(inferenceStartTime).withEndTime(inferenceEndTime) }
			).withAdditionalSignals(booleanSignals);
		
		final DtoTimeseriesFilesReaderConfig inferenceFilesReaderConfig =
			new DtoTimeseriesFilesReaderConfig(folderConfig.getInputFolder()).withFilesFilter(inferenceFilesFilter)
				.withDataFilter(inferenceDataFilter)
				.withDataTypes(dataTypes);
		
		final StateDetectionData inferenceData = StateDetectionData.create();
		inferenceData.readFiles(inferenceFilesReaderConfig);
		
		System.out.println("Creating inference ...");
		// build inference config - for more options here, see documentation
		// skipOnInsufficientData: if inference data does not meet the inference data specification for a timestamp an inference is demanded ("insufficient data"),
		// no exception is thrown but it's simply "skipped"
		final DtoInferenceConfig inferenceConfig = new DtoInferenceConfig(true);
		// note that we use here try-with-resource notation to destroy inference handle, inference.destroy() works equivalently
		try (final StateDetectionInference inference = StateDetectionInference.createByTraining(training, inferenceConfig)) {
			
			training.destroy();
			System.out.println("\t... done!");
			
			System.out.println("Calculating scores ...");
			// here, we choose exactly the timestamp that are in the inference data file
			final List<Long> inferenceTimestamps = new ArrayList<Long>();
			for (long time = inferenceStartTime; time < inferenceEndTime; time += 3600000L) {
				inferenceTimestamps.add(time);
			}
			final List<? extends IDtoSegmentsFloatDataPoint> scores = inference.infer(inferenceData, inferenceTimestamps);
			System.out.println("\t... done!");
			
			System.out.println("Saving inference files ...");
			Utils.writeSegmentTimeseries(new File(folderConfig.getOutputFolder(), "scores.json"), scores);
			Utils.writeAdditionalSignalsTimeseriesWithTarget(
				new File(folderConfig.getOutputFolder(), "additional-signals.json"),
				inferenceData.getBooleanSignal(TARGET_SIGNAL_ID),
				TARGET_SIGNAL_ID
			);
			inferenceData.destroy();
			System.out.println("\t... done!");
			
		}
	}
}
