package com.vernaio.example;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.fasterxml.jackson.datatype.jdk8.Jdk8Module;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoBooleanDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoFloatDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoBooleanDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoSegmentFloatValue;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoSegmentsFloatDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.exception.EngineException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.MissingLibraryException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.WrongFlavourException;
import com.vernaio.aivis.engine.v2.state.detection.sdk.StateDetectionData;

public class Utils {
	
	private static final ObjectMapper MAPPER = new ObjectMapper().registerModule(new Jdk8Module());
	
	public static abstract class Signal {
		
		private final String signalId;
		
		protected Signal(final String signalId) {
			this.signalId = signalId;
		}
		
		public String getSignalId() {
			return this.signalId;
		}
	}
	
	public static class FloatSignal extends Signal {
		
		private final List<DtoFloatDataPoint> dataPoints;
		
		protected FloatSignal(final String signalId) {
			super(signalId);
			this.dataPoints = new ArrayList<DtoFloatDataPoint>();
		}
		
		public List<DtoFloatDataPoint> getDataPoints() {
			return this.dataPoints;
		}
	}
	
	public static class BooleanSignal extends Signal {
		
		private final List<DtoBooleanDataPoint> dataPoints;
		
		protected BooleanSignal(final String signalId) {
			super(signalId);
			this.dataPoints = new ArrayList<DtoBooleanDataPoint>();
		}
		
		public List<DtoBooleanDataPoint> getDataPoints() {
			return this.dataPoints;
		}
	}
	
	public static <T> T readJson(final File file, final Class<T> clazz) throws JsonParseException, JsonMappingException, IOException {
		System.out.println(String.format("\t... reading file '%s' ...", file.getPath()));
		return MAPPER.readValue(file, clazz);
	}
	
	public static void writeAdditionalSignalsTimeseriesWithTarget(
		final File file,
		final List<? extends IDtoBooleanDataPoint> targetDataPoints,
		final String targetSignalId
	) throws FileNotFoundException, IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		
		final ArrayNode json = MAPPER.createArrayNode();
		final ObjectNode timeseriesByIdNode = MAPPER.createObjectNode();
		json.add(timeseriesByIdNode);
		
		timeseriesByIdNode.put("id", targetSignalId);
		
		final ObjectNode timeseriesNode = MAPPER.createObjectNode();
		timeseriesByIdNode.set("timeseries", timeseriesNode);
		
		timeseriesNode.put("_type", "Boolean");
		final ArrayNode dataPointsNode = timeseriesNode.putArray("dataPoints");
		for (IDtoBooleanDataPoint dataPoint : targetDataPoints) {
			final ObjectNode dataPointNode = MAPPER.createObjectNode();
			dataPointNode.put("timestamp", dataPoint.getTimestamp());
			dataPointNode.put("value", dataPoint.getValue());
			dataPointsNode.add(dataPointNode);
		}
		
		final FileWriter fileWriter = new FileWriter(file);
		fileWriter.write(json.toPrettyString());
		fileWriter.close();
		
		return;
		
	}
	
	public static void writeSegmentTimeseries(final File file, final List<? extends IDtoSegmentsFloatDataPoint> timeseries) throws FileNotFoundException,
		IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		
		final ObjectNode json = MAPPER.createObjectNode();
		
		json.put("_type", "SegmentsFloat");
		final ArrayNode dataPointsNode = json.putArray("dataPoints");
		for (IDtoSegmentsFloatDataPoint dataPoint : timeseries) {
			final ObjectNode dataPointNode = MAPPER.createObjectNode();
			dataPointNode.put("timestamp", dataPoint.getTimestamp());
			final ArrayNode valuesNode = dataPointNode.putArray("values");
			for (IDtoSegmentFloatValue segmentValue : dataPoint.getValues()) {
				final ObjectNode valueNode = MAPPER.createObjectNode();
				valueNode.put("segmentId", segmentValue.getSegmentId());
				valueNode.put("value", segmentValue.getValue());
				valuesNode.add(valueNode);
			}
			
			dataPointsNode.add(dataPointNode);
		}
		
		final FileWriter fileWriter = new FileWriter(file);
		fileWriter.write(json.toPrettyString());
		fileWriter.close();
		
	}
	
	public static <T> void writeJson(final File file, final T obj) throws JsonGenerationException, JsonMappingException, IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		Optional.ofNullable(file.getParentFile()).ifPresent(f -> f.mkdirs());
		MAPPER.writeValue(file, obj);
	}
	
}
