# this python script performs the same calculation as example_sd.ipynb

import json
import argparse
import os

from aivis_engine_v2_sd_sdk_python import (
    EngineLogger,
    StateDetectionSetup,
    StateDetectionData,
    StateDetectionAnalysis,
    StateDetectionTraining,
    StateDetectionInference,
)


# ---------------------------------------------------------------------------------
# Data location

TRAINING_DATA_FILE = "train_sd.csv"
INFERENCE_DATA_FILE = "eval_sd.csv"
TARGET_SIGNAL_ID = "failure"


# Input and output folder specified via args
def from_args():
    parser = argparse.ArgumentParser(
        description="Example project for signal prediction"
    )
    # data is expected in input folder
    parser.add_argument(
        "--input", "-i", type=str, nargs=1, required=True, help="Path to input folder"
    )
    # output files will be written to output folder
    parser.add_argument(
        "--output", "-o", type=str, nargs=1, required=True, help="Path to output folder"
    )
    args = parser.parse_args()
    # create output folder if not existing
    if not os.path.exists(args.output[0]):
        os.mkdir(args.output[0])
    return {
        "input_folder": args.input[0],
        "output_folder": args.output[0],
    }


# ---------------------------------------------------------------------------------
# Logger and Threads


class Logger(EngineLogger):
    def log(self, level, thread, module, message):
        min_logging_level = 3
        if level <= min_logging_level:
            print("\t... %s" % message)


def setup_logger_and_threads():
    print("Initializing StateDetectionSetup: logging and threads ...")
    StateDetectionSetup.register_logger(Logger())
    max_threads = 0  # number of threads will be chosen automatically with 0
    StateDetectionSetup.init_thread_count(max_threads)
    print("\t... done!")
    return


# ---------------------------------------------------------------------------------
# Reading and Writing


def read_json(filename):
    print("\t... reading file '%s' ..." % filename)
    with open(filename, "r") as file:
        return json.loads(file.read())


def write_json(filename, obj):
    print("\t... writing file '%s' ..." % filename)
    with open(filename, "w") as file:
        file.write(json.dumps(obj, indent=2))
    return


# convert DtoSegmentsFloatDataPoint's value to json
def dto_segments_float_point_to_json(values_object):
    return [
        {"segmentId": int(value_object.segment_id), "value": float(value_object.value)}
        for value_object in values_object
    ]


# convert [DtoSegmentsFloatDataPoint] to json
def segments_float_data_points_to_timeseries_json(list_object):
    return {
        "_type": "SegmentsFloat",
        "dataPoints": [
            {
                "timestamp": int(point_object.timestamp),
                "values": dto_segments_float_point_to_json(point_object.values),
            }
            for point_object in list_object
        ],
    }


# convert [DtoBooleanDataPoint] to json
def boolean_data_points_to_additional_signals_json(list_object):
    return [
        {
            "id": TARGET_SIGNAL_ID,
            "timeseries": {
                "_type": "Boolean",
                "dataPoints": [
                    {
                        "timestamp": int(point_object.timestamp),
                        "value": bool(point_object.value),
                    }
                    for point_object in list_object
                ],
            },
        }
    ]


# ---------------------------------------------------------------------------------
# MAIN


def main():
    folder_config = from_args()  # parsing args as folder_config
    setup_logger_and_threads()

    print("Loading training data ...")
    training_data = StateDetectionData.create()

    files_filter = {"includeFiles": [TRAINING_DATA_FILE]}
    data_types = {"booleanSignals": [TARGET_SIGNAL_ID]}
    files_reader_config = json.dumps(
        {
            "folder": folder_config["input_folder"],
            "filesFilter": files_filter,
            "dataTypes": data_types,
        }
    )

    training_data.read_files(files_reader_config)

    # print some signal to illustrate type DtoFloatDataPoint
    print("Example signal!")
    some_signal_id = "sensor_measurement_7"
    print("\t... Signal id:", some_signal_id)
    some_signal_data_points = training_data.get_float_signal(some_signal_id)
    print("\t... Signal length:", len(some_signal_data_points))
    print("\t... Signal entry number 42:", some_signal_data_points[42])
    print(
        "\t... Signal entry number 42 timestamp (unix/millisec):",
        some_signal_data_points[42].timestamp,
    )
    print("\t... Signal entry number 42 value:", some_signal_data_points[42].value)

    # ---- ANALYSIS ----

    print("Creating analysis ...")
    # build analysis config - for more options see documentation
    analysis_config = json.dumps(
        {
            "target": {
                "signal": TARGET_SIGNAL_ID  # an incident is if this signal switches to True
            },
            "selfLabeling": {
                "coherencePeriod": 432000000,  # how long before incidents aivis looks for state changes
                "mesh": 3600000,  # mesh for aivis to look for state changes in the coherence period
            },
            "sampling": {"additionalSampleMesh": None},
        }
    )
    # create and perform analysis
    analysis = StateDetectionAnalysis.create(training_data, analysis_config)
    print("\t... done!")

    # get preparation needed for training
    preparation = json.loads(analysis.get_training_preparation())

    print("Save analysis report file ...")
    analysis_report = json.loads(
        analysis.get_report()
    )  # contains analysis/segmentation infos
    write_json(
        os.path.join(folder_config["output_folder"], "analysis-report.json"),
        analysis_report,
    )
    print("\t... done!")

    # ---- TRAINING ----

    print("Creating training ...")
    # build training config - for more options see documentation
    training_config = json.dumps(
        {
            "sampling": {
                "maximalSampleCount": 5000,
            },
            "functions": preparation["functions"],
            "commonLabelHistory": preparation["commonLabelHistory"],
            "segments": preparation["zoomLevels"][0]["segments"],
        }
    )

    # create training and train the model
    training = StateDetectionTraining.create(training_data, training_config)
    print("\t... done!")

    print("Save training files ...")
    model = json.loads(training.get_model())
    write_json(os.path.join(folder_config["output_folder"], "model.json"), model)

    training_report = json.loads(training.get_report())  # training infos
    write_json(
        os.path.join(folder_config["output_folder"], "training-report.json"),
        training_report,
    )
    print("\t... done!")

    # ---- INFERENCE ----

    print("Loading inference data ...")
    inference_data = StateDetectionData.create()

    files_filter = {"includeFiles": [INFERENCE_DATA_FILE]}
    data_types = {"booleanSignals": [TARGET_SIGNAL_ID]}
    data_filter = {
        "_type": "Inference",
        "dataSpecifications": [model["inferenceDataSpecification"]],
        "additionalSignals": [TARGET_SIGNAL_ID],
        "inferenceTimeRanges": [{}],
    }
    files_reader_config = json.dumps(
        {
            "folder": folder_config["input_folder"],
            "filesFilter": files_filter,
            "dataFilter": data_filter,
            "dataTypes": data_types,
        }
    )

    inference_data.read_files(files_reader_config)

    print("Creating inference ...")
    inference_config = json.dumps({"skipOnInsufficientData": True})
    # via model is also an option: StateDetectionInference.create_by_model(model_json, inference_config)
    inference = StateDetectionInference.create_by_training(training, inference_config)
    print("\t... done!")

    print("Calculating inferences ...")
    # choose inference timestamps: selecting the timestamps from evaluation data,
    # but of course any timestamp where the inference data meets the model's inference data specification is possible
    timestamps = list(range(1328472000000, 1336572000000, 3600000))
    # infer scores
    scores = inference.infer(inference_data, timestamps)
    print("\t... done!")

    print("Save inference files ...")
    write_json(
        os.path.join(folder_config["output_folder"], "scores.json"),
        segments_float_data_points_to_timeseries_json(scores),
    )
    # write target from inference timespan to compare to scores later
    write_json(
        os.path.join(folder_config["output_folder"], "additional-signals.json"),
        boolean_data_points_to_additional_signals_json(
            inference_data.get_boolean_signal(TARGET_SIGNAL_ID)
        ),
    )
    print("\t... done!")

    print("Destroy everything ...")
    # destroying can be done once the respective object is not needed anymore - do this if memory is an issue
    training_data.destroy()
    analysis.destroy()
    training.destroy()
    inference_data.destroy()
    inference.destroy()
    print("\t... done!")

    print("Generating report ...")

    # ---- Toolbox HTML Report ----
    from aivis_engine_v2_toolbox.api import (
        build_state_detection_report,
        build_state_detection_analysis_report,
    )

    build_state_detection_analysis_report(
        {
            "title": "Aivis Engine V2 - State Detection Example: Machine Failures",
            "analysisReportFile": os.path.join(
                folder_config["output_folder"], "analysis-report.json"
            ),
            "archive": {
                "files": [
                    os.path.join(folder_config["input_folder"], TRAINING_DATA_FILE)
                ]
            },
            "outputFile": os.path.join(
                folder_config["output_folder"],
                "aivis-engine-v2-state-detection-example-analysis.html",
            ),
        }
    )

    build_state_detection_report(
        {
            "title": "Aivis Engine V2 - State Detection Example: Machine Failures",
            "scores": {
                "file": os.path.join(folder_config["output_folder"], "scores.json")
            },
            "target": {
                "file": os.path.join(
                    folder_config["output_folder"], "additional-signals.json"
                )
            },
            "analysisReportFile": os.path.join(
                folder_config["output_folder"], "analysis-report.json"
            ),
            "trainingReportFile": os.path.join(
                folder_config["output_folder"], "training-report.json"
            ),
            "outputFile": os.path.join(
                folder_config["output_folder"],
                "aivis-engine-v2-state-detection-example-training.html",
            ),
        }
    )

    print("ALL DONE!")
    return


# run main function
main()
